from __future__ import annotations

from tumfl.AST.Expression.Expression import Expression
from tumfl.Token import Token

from .Statement import Statement


class Block(Statement):
    """A block consisting of statements"""

    def __init__(
        self,
        token: Token,
        statements: list[Statement],
        returns: list[Expression],
    ):
        super().__init__(token, "Block")
        self.statements: list[Statement] = statements
        self.returns: list[Expression] = returns



from __future__ import annotations

from abc import ABC

from tumfl.AST.ASTNode import ASTNode
from tumfl.Token import Token


class Statement(ASTNode, ABC):
    """Baseclass for all statements, must contain comments"""

    def __init__(self, token: Token, name: str):
        super().__init__(token, name)
        self.comment: list[str] = token.comment



from __future__ import annotations

from tumfl.Token import Token, TokenType

from .Expression import Expression


class Boolean(Expression):
    """A boolean value (true or false)"""

    def __init__(self, token: Token, value: bool) -> None:
        super().__init__(token, "Boolean")
        self.value: bool = value

    @staticmethod
    def from_token(token: Token) -> Boolean:
        assert token.type in [TokenType.TRUE, TokenType.FALSE]
        return Boolean(token, token.type == TokenType.TRUE)



from __future__ import annotations

import string
from enum import Enum
from typing import Iterator, Optional, Sequence, Type

from .AST import *
from .basic_walker import BasicWalker


class FormattingStyle:
    # pylint: disable=too-few-public-methods

    # Statement separator
    STATEMENT_SEPARATOR: str = "\n"
    # Indentation character
    INDENTATION: str = "\t"
    # Argument separator
    ARGUMENT_SEPARATOR: str = ", "
    # Include comments or omit them
    INCLUDE_COMMENTS: bool = True
    # Separator after --
    COMMENT_SEP: str = " "
    # Use single quotes if this saves on string escapes
    USE_SINGLE_QUOTE: bool = False
    # Use direct table and string calls
    USE_CALL_SHORTHAND: bool = False
    # Remove all characters that are not strictly required
    REMOVE_UNNECESSARY_CHARS: bool = False
    # Add brackets to arithmetic expressions no matter the use
    ADD_ALL_BRACKETS: bool = False
    # Add brackets to arithmetic expressions if the order is potentially confusing
    ADD_CLOSE_BRACKETS: bool = True


class MinifiedStyle(FormattingStyle):
    # pylint: disable=too-few-public-methods

    STATEMENT_SEPARATOR = ";"
    INDENTATION = ""
    ARGUMENT_SEPARATOR = ","
    INCLUDE_COMMENTS = False
    COMMENT_SEP = ""
    USE_SINGLE_QUOTE = True
    USE_CALL_SHORTHAND = True
    REMOVE_UNNECESSARY_CHARS = True


class Separators(Enum):
    Statement: str = "stmt"
    Newline: str = "newline"
    Argument: str = "arg"
    Space: str = " "
    Dot: str = "."
    Indent: str = "indent"
    DeIndent: str = "de-indent"

    def join(self, to_join: Iterator[Retype]) -> Retype:
        join_list: list[Retype] = list(to_join)
        for i in range(len(join_list) - 1, 0, -1):
            join_list[i:i] = [[self]]
        return [elem for part in join_list for elem in part]


Retype = list[str | Separators]


class Formatter(BasicWalker[Retype]):
    # pylint: disable=too-many-public-methods

    def __init__(self, style: Type[FormattingStyle]):
        self.s: Type[FormattingStyle] = style

    def _format_args(self, arguments: Sequence[ASTNode]) -> Retype:
        return Separators.Argument.join(self.visit(arg) for arg in arguments)

    @staticmethod
    def _find_level(to_check: str) -> int:
        level: int = 0
        while True:
            if f"[{'=' * level}[" not in to_check:
                break
            level += 1
        return level

    def _format_comment(self, comment: str, force_long: bool = False) -> Retype:
        comment = comment.strip()
        if force_long or "\n" in comment:
            level: int = self._find_level(comment)
            return [f"--[{'=' * level}[{comment}]{'=' * level}]", Separators.Statement]
        return [f"--{self.s.COMMENT_SEP}{comment}", Separators.Newline]

    def _format_function_args(self, args: Sequence[Expression]) -> Retype:
        if self.s.USE_CALL_SHORTHAND and len(args) == 1:
            if isinstance(args[0], String):
                return self.visit(args[0])
            if isinstance(args[0], Table):
                return self.visit(args[0])
        return ["(", *self._format_args(args), ")"]

    def visit_Assign(self, node: Assign) -> Retype:
        return [
            *self._format_args(node.targets),
            Separators.Space,
            "=",
            Separators.Space,
            *self._format_args(node.expressions),
        ]

    def visit_Block(self, node: Block) -> Retype:
        result: Retype = [Separators.Indent]
        for statement in node.statements:
            if self.s.INCLUDE_COMMENTS and statement.comment:
                for comment in statement.comment:
                    result += self._format_comment(comment)
            result += self.visit(statement) + [Separators.Statement]
        if node.returns:
            returns: Retype = self._format_args(node.returns)
            result += ["return", Separators.Space, *returns, Separators.Statement]
        result.append(Separators.DeIndent)
        return result

    def visit_Boolean(self, node: Boolean) -> Retype:
        return [str(node.value).lower()]

    def visit_Break(self, _node: Break) -> Retype:
        return ["break"]

    def visit_Chunk(self, node: Chunk) -> Retype:
        result: Retype = self.visit_Block(node)
        return [Separators.DeIndent, *result[:-2]]

    def visit_Goto(self, node: Goto) -> Retype:
        return ["goto", Separators.Space, *self.visit(node.label_name)]

    def visit_If(self, node: If) -> Retype:
        spaced_test: Retype = [
            Separators.Space,
            *self.visit(node.test),
            Separators.Space,
        ]
        result: Retype = ["if", *spaced_test, "then", Separators.Statement]
        result += self.visit(node.true)
        current_if: If = node
        while isinstance(current_if.false, If):
            current_if = current_if.false
            spaced_test = [
                Separators.Space,
                *self.visit(current_if.test),
                Separators.Space,
            ]
            result += ["elseif", *spaced_test, "then", Separators.Statement]
            result += self.visit(current_if.true)
        if current_if.false:
            result += ["else", Separators.Statement]
            result += self.visit(current_if.false)
        result += ["end"]
        return result

    def visit_Index(self, node: Index) -> Retype:
        return [*self.visit(node.lhs), "[", *self.visit(node.variable_name), "]"]

    def visit_Label(self, node: Label) -> Retype:
        return ["::", self.visit(node.label_name), "::"]

    def visit_Name(self, node: Name) -> Retype:
        return [node.variable_name]

    def visit_Nil(self, _node: Nil) -> Retype:
        return ["nil"]

    def visit_Number(self, node: Number) -> Retype:
        return [str(node)]

    def visit_Repeat(self, node: Repeat) -> Retype:
        return [
            "repeat",
            Separators.Statement,
            *self.visit(node.body),
            "until",
            Separators.Space,
            *self.visit(node.condition),
        ]

    def visit_Semicolon(self, _node: Semicolon) -> Retype:
        return [";"]

    def visit_String(self, node: String) -> Retype:
        if "\n" in node.value:
            level: int = self._find_level(node.value)
            return [f"[{'=' * level}[{node.value}]{'=' * level}]"]
        if self.s.USE_SINGLE_QUOTE and node.value.count("'") < node.value.count('"'):
            return ["'" + node.value.replace('"', '\\"') + "'"]
        return ['"' + node.value.replace("'", "\\'") + '"']

    def visit_Table(self, node: Table) -> Retype:
        return ["{", *self._format_args(node.fields), "}"]

    def visit_Vararg(self, _node: Vararg) -> Retype:
        return ["..."]

    def visit_While(self, node: While) -> Retype:
        return [
            "while",
            Separators.Space,
            *self.visit(node.condition),
            Separators.Space,
            "do",
            Separators.Statement,
            *self.visit(node.body),
            "end",
        ]

    def _need_brackets(
        self, own_node: BinOp, other_node: Expression, care_unop: bool
    ) -> bool:
        own_precedence: int = own_node.op.get_precedence()
        close_precedence: int = own_precedence + (
            1 if self.s.ADD_CLOSE_BRACKETS else -1
        )
        return (
            self.s.ADD_ALL_BRACKETS
            or isinstance(other_node, BinOp)
            and (
                other_node.op.get_precedence() < own_precedence
                or other_node.op.get_precedence() == close_precedence
            )
            or (care_unop or self.s.ADD_CLOSE_BRACKETS)
            and isinstance(other_node, UnOp)
            and (10 < own_precedence or 10 == close_precedence)
        )

    def visit_BinOp(self, node: BinOp) -> Retype:
        result: Retype = []
        if self._need_brackets(node, node.left, True):
            result += ["(", *self.visit(node.left), ")"]
        else:
            result += self.visit(node.left)
        result += [Separators.Space, node.op.value, Separators.Space]
        if self._need_brackets(node, node.right, False):
            result += ["(", *self.visit(node.right), ")"]
        else:
            result += self.visit(node.right)
        return result

    def visit_FunctionCall(self, node: FunctionCall) -> Retype:
        return self.visit(node.function) + self._format_function_args(node.arguments)

    def visit_FunctionDefinition(self, node: FunctionDefinition) -> Retype:
        return [
            "function",
            Separators.Space,
            *Separators.Dot.join(self.visit(name) for name in node.names),
            "(",
            *self._format_args(node.parameters),
            ")",
            Separators.Statement,
            *self.visit(node.body),
            "end",
            Separators.Statement,
        ]

    def visit_IterativeFor(self, node: IterativeFor) -> Retype:
        return [
            "for",
            Separators.Space,
            *self._format_args(node.namelist),
            Separators.Space,
            "in",
            Separators.Space,
            *self._format_args(node.explist),
            Separators.Space,
            "do",
            Separators.Statement,
            *self.visit(node.body),
            "end",
        ]

    def visit_LocalAssign(self, node: LocalAssign) -> Retype:
        result: Retype = ["local", Separators.Space]
        result += Separators.Argument.join([str(arg)] for arg in node.variable_names)
        if node.expressions:
            result += [Separators.Space, "=", Separators.Space]
            result += self._format_args(node.expressions)
        return result

    def visit_MethodInvocation(self, node: MethodInvocation) -> Retype:
        return [
            *self.visit(node.function),
            ":",
            *self.visit(node.method),
            *self._format_function_args(node.arguments),
        ]

    def visit_NamedIndex(self, node: NamedIndex) -> Retype:
        return [*self.visit(node.lhs), ".", *self.visit(node.variable_name)]

    def visit_NumericFor(self, node: NumericFor) -> Retype:
        spaced_name: Retype = [
            Separators.Space,
            self.visit(node.variable_name),
            Separators.Space,
        ]
        result: Retype = ["for", *spaced_name, "=", Separators.Space]
        result += self.visit(node.start)
        result += [",", Separators.Space, *self.visit(node.stop)]
        if node.step:
            result += [",", Separators.Space, *self.visit(node.step)]
        result += [Separators.Space, "do", Separators.Statement]
        result += self.visit(node.body)
        result += ["end"]
        return result

    def visit_UnOp(self, node: UnOp) -> Retype:
        result: Retype = [node.op.value]
        if (
            self.s.ADD_ALL_BRACKETS
            or isinstance(node.right, BinOp)
            and node.right.op != BinaryOperand.EXPONENT
        ):
            result += ["(", *self.visit(node.right), ")"]
        elif isinstance(node.right, UnOp) or node.op == UnaryOperand.NOT:
            result += [Separators.Space, *self.visit(node.right)]
        else:
            result += self.visit(node.right)
        return result

    def visit_ExpFunctionCall(self, node: ExpFunctionCall) -> Retype:
        return self.visit(node.function) + self._format_function_args(node.arguments)

    def visit_ExpFunctionDefinition(self, node: ExpFunctionDefinition) -> Retype:
        return [
            "function",
            Separators.Space,
            "(",
            *self._format_args(node.parameters),
            ")",
            Separators.Statement,
            *self.visit(node.body),
            "end",
        ]

    def visit_ExpMethodInvocation(self, node: ExpMethodInvocation) -> Retype:
        return [
            *self.visit(node.function),
            ":",
            *self.visit(node.method),
            *self._format_function_args(node.arguments),
        ]

    def visit_LocalFunctionDefinition(self, node: LocalFunctionDefinition) -> Retype:
        return [
            "local",
            "function",
            Separators.Space,
            *self.visit(node.function_name),
            "(",
            *self._format_args(node.parameters),
            ")",
            Separators.Statement,
            *self.visit(node.body),
            "end",
            Separators.Statement,
        ]

    def visit_ExplicitTableField(self, node: ExplicitTableField) -> Retype:
        return [
            "[",
            *self.visit(node.at),
            "]",
            Separators.Space,
            "=",
            Separators.Space,
            *self.visit(node.value),
        ]

    def visit_NamedTableField(self, node: NamedTableField) -> Retype:
        return [
            *self.visit(node.field_name),
            Separators.Space,
            "=",
            Separators.Space,
            *self.visit(node.value),
        ]

    def visit_NumberedTableField(self, node: NumberedTableField) -> Retype:
        return self.visit(node.value)


def _sep_required(first: str, second: str) -> bool:
    return (
        first in string.ascii_letters
        and second in string.ascii_letters
        or first in string.digits
        and second in string.ascii_letters
        or first in string.ascii_letters
        and second in string.digits
        or first == "-"
        and second == "-"
    )


def _search_token(start_idx: int, direction: int, tokens: Retype) -> str | Separators:
    while isinstance(token := tokens[start_idx], Separators) and token in (
        Separators.Indent,
        Separators.DeIndent,
    ):
        start_idx += direction
    return token


def _remove_separators(token_stream: Retype) -> None:
    token_stream.append("/")
    for i in range(len(token_stream) - 2, 0, -1):
        if isinstance(own_token := token_stream[i], Separators):
            if own_token not in (Separators.Space, Separators.Statement):
                continue
            previous_token = _search_token(i - 1, -1, token_stream)
            next_token = _search_token(i + 1, 1, token_stream)
            if (
                isinstance(previous_token, Separators)
                or isinstance(next_token, Separators)
                or not _sep_required(previous_token[-1], next_token[0])
            ):
                token_stream.pop(i)
    token_stream.pop()


def _resolve_tokens(token_stream: Retype, style: Type[FormattingStyle]) -> None:
    newline: str = (
        style.STATEMENT_SEPARATOR if "\n" in style.STATEMENT_SEPARATOR else "\n"
    )
    for i, token in enumerate(token_stream):
        if isinstance(token, Separators):
            if token == Separators.Space:
                token_stream[i] = " "
            elif token == Separators.Dot:
                token_stream[i] = "."
            elif token == Separators.Statement:
                token_stream[i] = style.STATEMENT_SEPARATOR
            elif token == Separators.Argument:
                token_stream[i] = style.ARGUMENT_SEPARATOR
            elif token == Separators.Newline:
                token_stream[i] = newline


def _indent(token_stream: Retype, indentation_str: str) -> None:
    indentation_level: int = 0
    indentation_dirty: bool = False
    for i, token in enumerate(token_stream):
        if isinstance(token, Separators):
            assert token in (Separators.Indent, Separators.DeIndent)
            if token == Separators.Indent:
                indentation_level += 1
            else:
                indentation_level -= 1
        else:
            if indentation_dirty:
                token_stream[i] = indentation_level * indentation_str + token
                indentation_dirty = False
            if "\n" in token:
                indentation_dirty = True
    assert indentation_level == 0


def _join_tokens(token_stream: Retype) -> str:
    result: str = ""
    for token in token_stream:
        if isinstance(token, str):
            result += token
    return result


def format(ast: ASTNode, style: Optional[Type[FormattingStyle]] = None) -> str:
    style = style or FormattingStyle
    formatter: Formatter = Formatter(style)
    token_stream: Retype = formatter.visit(ast)
    if style.REMOVE_UNNECESSARY_CHARS:
        _remove_separators(token_stream)
    token_stream[0:0] = ["--tumfl", Separators.Newline]
    _resolve_tokens(token_stream, style)
    _indent(token_stream, style.INDENTATION)
    return _join_tokens(token_stream)



from __future__ import annotations

from typing import Optional

from tumfl.AST.BaseFunctionDefinition import BaseFunctionDefinition
from tumfl.AST.Expression.Name import Name
from tumfl.AST.Expression.Vararg import Vararg
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class FunctionDefinition(Statement):
    """Definition of a lua function, NOT local"""

    def __init__(
        self,
        token: Token,
        names: list[Name],
        method_name: Optional[Name],
        parameters: list[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "FunctionDefinition")
        self.names: list[Name] = names
        self.method_name: Optional[Name] = method_name
        self.parameters: list[Name | Vararg] = parameters
        self.body: Block = body

    @staticmethod
    def from_base_definition(
        base: BaseFunctionDefinition,
        names: list[Name],
        method_name: Optional[Name],
    ) -> FunctionDefinition:
        return FunctionDefinition(
            base.token, names, method_name, base.parameters, base.body
        )



import unittest

from tumfl.AST import Block, If
from tumfl.formatter import Formatter, FormattingStyle, MinifiedStyle, Separators
from tumfl.parser import Parser


class TestFormatter(unittest.TestCase):
    def setUp(self) -> None:
        self.normal: Formatter = Formatter(FormattingStyle)
        self.minified: Formatter = Formatter(MinifiedStyle)

    def test_format_comment(self):
        expected = ["-- comment", Separators.Newline]
        self.assertEqual(self.normal._format_comment("comment"), expected)
        expected = ["--comment", Separators.Newline]
        self.assertEqual(self.minified._format_comment("comment"), expected)
        expected = ["--[[ab\ncd]]", Separators.Statement]
        self.assertEqual(self.normal._format_comment("ab\ncd"), expected)
        expected = ["--[==[[[=[\na]==]", Separators.Statement]
        self.assertEqual(self.normal._format_comment("[[=[\na"), expected)

    def test_format_function_args(self):
        parser = Parser("'abc'{}1'ghi','jkl'")
        expression = parser._parse_exp()  # "abc"
        expected = ["(", *self.normal.visit(expression), ")"]
        self.assertEqual(self.normal._format_function_args([expression]), expected)
        expected = self.normal.visit(expression)
        self.assertEqual(self.minified._format_function_args([expression]), expected)
        expression = parser._parse_exp()  # {}
        expected = ["(", *self.normal.visit(expression), ")"]
        self.assertEqual(self.normal._format_function_args([expression]), expected)
        expected = self.normal.visit(expression)
        self.assertEqual(self.minified._format_function_args([expression]), expected)
        expression = parser._parse_exp()  # 1
        expected = ["(", *self.normal.visit(expression), ")"]
        self.assertEqual(self.minified._format_function_args([expression]), expected)
        expression = parser._parse_exp_list()  # "ghi", "jkl"
        expected = [
            "(",
            *self.normal.visit(expression[0]),
            Separators.Argument,
            *self.normal.visit(expression[1]),
            ")",
        ]
        self.assertEqual(self.normal._format_function_args(expression), expected)
        self.assertEqual(self.minified._format_function_args(expression), expected)

    def test_Assign(self):
        stmt = Parser("a = b")._parse_statement()
        expected = ["a", Separators.Space, "=", Separators.Space, "b"]
        self.assertEqual(self.normal.visit(stmt), expected)
        self.assertEqual(self.minified.visit(stmt), expected)

    def test_block(self):
        chunk = Parser("--comment\na=b c=d return a").parse_chunk()
        expected = [
            Separators.DeIndent,
            Separators.Indent,
            "-- comment",
            Separators.Newline,
            *self.normal.visit(chunk.statements[0]),
            Separators.Statement,
            *self.normal.visit(chunk.statements[1]),
            Separators.Statement,
            "return",
            Separators.Space,
            "a",
        ]
        self.assertEqual(self.normal.visit(chunk), expected)
        expected.pop(2)
        expected.pop(2)
        self.assertEqual(self.minified.visit(chunk), expected)
        chunk = Parser("a=b").parse_chunk()
        expected = [
            Separators.DeIndent,
            Separators.Indent,
            *self.normal.visit(chunk.statements[0]),
        ]
        self.assertEqual(self.normal.visit(chunk), expected)

    def test_Boolean(self):
        parser = Parser("true false")
        expr = parser._parse_exp()
        expected = ["true"]
        self.assertEqual(self.normal.visit(expr), expected)
        self.assertEqual(self.minified.visit(expr), expected)
        expr = parser._parse_exp()
        expected = ["false"]
        self.assertEqual(self.normal.visit(expr), expected)
        self.assertEqual(self.minified.visit(expr), expected)

    def test_Break(self):
        expr = Parser("break")._parse_break()
        expected = ["break"]
        self.assertEqual(self.normal.visit(expr), expected)

    def test_Goto(self):
        stmt = Parser("goto label")._parse_statement()
        expected = ["goto", Separators.Space, "label"]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_If(self):
        stmt = Parser("if 1 then a=b elseif 0 then c=d else e=f end")._parse_if()
        assert isinstance(stmt.false, If)
        assert isinstance(stmt.false.false, Block)
        expected = [
            "if",
            Separators.Space,
            "1",
            Separators.Space,
            "then",
            Separators.Statement,
            *self.normal.visit(stmt.true),
            "elseif",
            Separators.Space,
            "0",
            Separators.Space,
            "then",
            Separators.Statement,
            *self.normal.visit(stmt.false.true),
            "else",
            Separators.Statement,
            *self.normal.visit(stmt.false.false),
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)
        stmt = Parser("if 1 then a=b end")._parse_if()
        expected = [
            "if",
            Separators.Space,
            "1",
            Separators.Space,
            "then",
            Separators.Statement,
            *self.normal.visit(stmt.true),
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)



from .BinOp import BinaryOperand, BinOp
from .Boolean import Boolean
from .ExpFunctionCall import ExpFunctionCall
from .ExpFunctionDefinition import ExpFunctionDefinition
from .ExpMethodInvocation import ExpMethodInvocation
from .Expression import Expression
from .Index import Index
from .Name import Name
from .NamedIndex import NamedIndex
from .Nil import Nil
from .Number import Number
from .String import String
from .Table import Table
from .TableField import (
    ExplicitTableField,
    NamedTableField,
    NumberedTableField,
    TableField,
)
from .UnOp import UnaryOperand, UnOp
from .Vararg import Vararg
from .Variable import Variable



from __future__ import annotations

from tumfl.Token import Token

from .Expression import Expression
from .TableField import TableField


class Table(Expression):
    """A lua table constructor, like {[1]=2, 3; b=4}"""

    def __init__(self, token: Token, fields: list[TableField]):
        super().__init__(token, "Table")
        self.fields: list[TableField] = fields



from __future__ import annotations

from tumfl.Token import Token, TokenType

from .Expression import Expression


class Vararg(Expression):
    """The vararg expression (...)"""

    def __init__(self, token: Token) -> None:
        super().__init__(token, "Vararg")

    @staticmethod
    def from_token(token: Token) -> Vararg:
        assert token.type == TokenType.ELLIPSIS
        return Vararg(token)



from __future__ import annotations

from tumfl.AST.Expression.Expression import Expression
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class While(Statement):
    """The while loop"""

    def __init__(self, token: Token, condition: Expression, body: Block):
        super().__init__(token, "While")
        self.condition: Expression = condition
        self.body: Block = body



from __future__ import annotations

from typing import Optional

from tumfl.AST.Expression.Expression import Expression
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class If(Statement):
    """An if statement, may contain a block or other if as false condition"""

    def __init__(
        self,
        token: Token,
        test: Expression,
        true: Block,
        false: Optional[Block | If],
    ):
        super().__init__(token, "If")
        self.test: Expression = test
        self.true: Block = true
        self.false: Optional[Block | If] = false



from __future__ import annotations

from tumfl.AST.Expression.Expression import Expression
from tumfl.Token import Token

from .Statement import Statement


class Assign(Statement):
    """The assignment statement, like a,b = 1,nil"""

    def __init__(
        self,
        token: Token,
        targets: list[Expression],
        expressions: list[Expression],
    ):
        super().__init__(token, "Assign")
        self.targets: list[Expression] = targets
        self.expressions: list[Expression] = expressions



from __future__ import annotations

from typing import TYPE_CHECKING

from .Token import Token

if TYPE_CHECKING:
    from .parser import Hint


class TumflException(Exception):
    ...


class ParserException(TumflException):
    def __init__(self, message: str, hints: list[Hint], token: Token):
        super().__init__(message)
        self.hints: list[Hint] = hints
        self.token: Token = token


class LexerException(TumflException):
    def __init__(self, message: str, line: int, column: int):
        super().__init__(message)
        self.line: int = line
        self.column: int = column



from inspect import Signature, signature
from typing import Any, List, Optional


def generic_str(
    self: Any,
    ignored_keywords: Optional[List[str]] = None,
    explicit_keywords: Optional[List[str]] = None,
) -> str:
    newlines: bool = True
    newline_or_space: str = "\n" if newlines else " "
    newline_or_nothing: str = "\n" if newlines else ""
    string: str = f"{self.__class__.__name__}({newline_or_nothing}"
    dir: List[str] = []
    self_signature: Signature = signature(self.__init__)
    ignored_keywords = (ignored_keywords or []) + ["self"]
    for i in self_signature.parameters:
        if i not in ignored_keywords:
            dir.append(i)
    if explicit_keywords:
        for i in explicit_keywords:
            if i not in self_signature.parameters:
                dir.append(i)
    if dir:
        for i in dir:
            string += f'{"    " if newlines else ""}{i}='
            lines = repr(getattr(self, i)).splitlines()
            string += f"{lines.pop(0)}{newline_or_space}"
            for j in lines:
                string += f'{"    "}{j}{newline_or_space}'
            string = f"{string[:-1]},{newline_or_space}"
        string = string[:-1]
    return f"{string[:-1]}{newline_or_nothing})"



from __future__ import annotations

from tumfl.Token import Token, TokenType

from .Expression import Expression


class Nil(Expression):
    """The nil value"""

    def __init__(self, token: Token) -> None:
        super().__init__(token, "Nil")

    @staticmethod
    def from_token(token: Token) -> Nil:
        assert token.type == TokenType.NIL
        return Nil(token)



from __future__ import annotations

from typing import Any, Optional

from tumfl.AST.Expression.Expression import Expression
from tumfl.AST.Expression.Name import Name
from tumfl.Token import Token

from .Statement import Statement


class AttributedName:
    def __init__(self, name: Name, attribute: Optional[Name] = None):
        self.name: Name = name
        self.attribute: Optional[Name] = attribute

    def __repr__(self) -> str:
        return f"AttributedName(name={self.name!r}, attribute={self.attribute!r})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AttributedName):
            return self.name == other.name and self.attribute == other.attribute
        return False

    def __str__(self) -> str:
        if self.attribute:
            return f"{self.name.variable_name} <{self.attribute.variable_name}>"
        return self.name.variable_name


class LocalAssign(Statement):
    """Assignment of local variables"""

    def __init__(
        self,
        token: Token,
        variable_names: list[AttributedName],
        expressions: Optional[list[Expression]],
    ):
        super().__init__(token, "LocalAssign")
        self.variable_names: list[AttributedName] = variable_names
        self.expressions: Optional[list[Expression]] = expressions



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from tumfl.Token import Token

from .ASTNode import ASTNode

if TYPE_CHECKING:
    from .Expression.Name import Name
    from .Expression.Vararg import Vararg
    from .Statement.Block import Block


class BaseFunctionDefinition(ASTNode, ABC):
    """Definition of a lua function, NOT local"""

    def __init__(
        self,
        token: Token,
        parameters: list[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "BaseFunctionDefinition")
        self.parameters: list[Name | Vararg] = parameters
        self.body: Block = body



from __future__ import annotations

from tumfl.Token import Token

from .Expression import Expression
from .Variable import Variable


class Index(Variable):
    """A index, like a[1], b[c] or d["abc"]"""

    def __init__(self, token: Token, lhs: Expression, variable_name: Expression):
        super().__init__(token, "Index")
        self.lhs: Expression = lhs
        self.variable_name: Expression = variable_name



from __future__ import annotations

from tumfl.AST.BaseFunctionDefinition import BaseFunctionDefinition
from tumfl.AST.Statement.Block import Block
from tumfl.Token import Token

from .Expression import Expression
from .Name import Name
from .Vararg import Vararg


class ExpFunctionDefinition(Expression):
    """Definition of a lua function expression"""

    def __init__(
        self,
        token: Token,
        parameters: list[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "ExpFunctionDefinition")
        self.parameters: list[Name | Vararg] = parameters
        self.body: Block = body

    @staticmethod
    def from_base_definition(base: BaseFunctionDefinition) -> ExpFunctionDefinition:
        return ExpFunctionDefinition(base.token, base.parameters, base.body)



from __future__ import annotations

from tumfl.AST.Expression.Name import Name
from tumfl.Token import Token

from .Statement import Statement


class Goto(Statement):
    """The goto statement"""

    def __init__(self, token: Token, label_name: Name):
        super().__init__(token, "Goto")
        self.label_name: Name = label_name



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, NoReturn, TypeVar

from .AST import *

T = TypeVar("T")


class NodeVisitor(Generic[T]):
    def visit(self, node: ASTNode) -> T:
        method_name = "visit_" + type(node).__name__
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node: ASTNode) -> NoReturn:
        raise RuntimeError(f"No visit_{type(node).__name__} method")


class BasicWalker(NodeVisitor, Generic[T], ABC):
    # pylint: disable=unused-argument,too-many-public-methods
    def visit_BinOp(self, node: BinOp) -> T:
        self.visit(node.right)
        return self.visit(node.left)

    @abstractmethod
    def visit_Boolean(self, node: Boolean) -> T:
        raise NotImplementedError

    def visit_ExpFunctionCall(self, node: ExpFunctionCall) -> T:
        ret: T = self.visit(node.function)
        for argument in node.arguments:
            self.visit(argument)
        return ret

    def visit_ExpFunctionDefinition(self, node: ExpFunctionDefinition) -> T:
        for parameter in node.parameters:
            self.visit(parameter)
        return self.visit(node.body)

    def visit_ExpMethodInvocation(self, node: ExpMethodInvocation) -> T:
        self.visit(node.function)
        ret: T = self.visit(node.method)
        for argument in node.arguments:
            self.visit(argument)
        return ret

    def visit_Index(self, node: Index) -> T:
        self.visit(node.lhs)
        return self.visit(node.variable_name)

    @abstractmethod
    def visit_Name(self, node: Name) -> T:
        raise NotImplementedError

    def visit_NamedIndex(self, node: NamedIndex) -> T:
        self.visit(node.lhs)
        return self.visit(node.variable_name)

    @abstractmethod
    def visit_Nil(self, node: Nil) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_Number(self, node: Number) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_String(self, node: String) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_Table(self, node: Table) -> T:
        raise NotImplementedError

    def visit_UnOp(self, node: UnOp) -> T:
        return self.visit(node.right)

    @abstractmethod
    def visit_Vararg(self, node: Vararg) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_Assign(self, node: Assign) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_Block(self, node: Block) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_Break(self, node: Break) -> T:
        raise NotImplementedError

    def visit_Chunk(self, node: Chunk) -> T:
        return self.visit_Block(node)

    def visit_FunctionCall(self, node: FunctionCall) -> T:
        ret: T = self.visit(node.function)
        for argument in node.arguments:
            self.visit(argument)
        return ret

    def visit_FunctionDefinition(self, node: FunctionDefinition) -> T:
        for name in node.names:
            self.visit(name)
        if node.method_name:
            self.visit(node.method_name)
        for parameter in node.parameters:
            self.visit(parameter)
        return self.visit(node.body)

    @abstractmethod
    def visit_Goto(self, node: Goto) -> T:
        raise NotImplementedError

    def visit_If(self, node: If) -> T:
        self.visit(node.test)
        ret: T = self.visit(node.true)
        if node.false:
            self.visit(node.false)
        return ret

    def visit_IterativeFor(self, node: IterativeFor) -> T:
        for name in node.namelist:
            self.visit(name)
        for expression in node.explist:
            self.visit(expression)
        return self.visit(node.body)

    @abstractmethod
    def visit_Label(self, node: Label) -> T:
        raise NotImplementedError

    @abstractmethod
    def visit_LocalAssign(self, node: LocalAssign) -> T:
        raise NotImplementedError

    def visit_LocalFunctionDefinition(self, node: LocalFunctionDefinition) -> T:
        self.visit(node.function_name)
        for parameter in node.parameters:
            self.visit(parameter)
        return self.visit(node.body)

    def visit_MethodInvocation(self, node: MethodInvocation) -> T:
        ret: T = self.visit(node.function)
        self.visit(node.method)
        for argument in node.arguments:
            self.visit(argument)
        return ret

    def visit_NumericFor(self, node: NumericFor) -> T:
        self.visit(node.variable_name)
        self.visit(node.start)
        self.visit(node.stop)
        if node.step:
            self.visit(node.step)
        return self.visit(node.body)

    def visit_Repeat(self, node: Repeat) -> T:
        self.visit(node.condition)
        return self.visit(node.body)

    @abstractmethod
    def visit_Semicolon(self, node: Semicolon) -> T:
        raise NotImplementedError

    def visit_While(self, node: While) -> T:
        self.visit(node.condition)
        return self.visit(node.body)

    def visit_NamedTableField(self, node: NamedTableField) -> T:
        self.visit(node.field_name)
        return self.visit(node.value)

    def visit_NumberedTableField(self, node: NumberedTableField) -> T:
        return self.visit(node.value)

    def visit_ExplicitTableField(self, node: ExplicitTableField) -> T:
        self.visit(node.at)
        return self.visit(node.value)
